/***************************************************************************
 *            timeline.c
 *
 *  Thu Jul 22 18:54:07 2004
 *  Copyright  2004  benjie
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <string.h>

#include "support.h"
#include "tocdoc.h"
#include "callbacks.h"
#include "interface.h"
#include "timeline.h"
#include "thumbnail.h"
#include "inspector.h"
#include "audio.h"

#define TL_PIXELS_PER_FRAME 64
#define TL_ZOOM_START (256 * TL_PIXELS_PER_FRAME)
#define TL_ZOOM_MIN 1
#define TL_ZOOM_MAX 40000
#define TL_ZOOM_INC 1.3
#define TL_ENTRY_TOP 45
#define TL_ENTRY_BOTTOM 5

GdkPixbuf   *markImg[2];
const gint rulerSize[] = {10, 35, FRAMES_PER_SEC, FRAMES_PER_SEC*2,
                          FRAMES_PER_SEC*5, FRAMES_PER_SEC*10,
                          FRAMES_PER_SEC*15, FRAMES_PER_SEC*30,
                          FRAMES_PER_SEC*60, FRAMES_PER_SEC*60*2,
                          FRAMES_PER_SEC*60*5, FRAMES_PER_SEC*60*10,
                          FRAMES_PER_SEC*60*15, FRAMES_PER_SEC*60*30, 0};


extern GtkStatusbar *statusBar;
extern guint statusBarID;
extern gint drawCursorTag;
extern prefsStruct  prefs;
extern GMutex *thumbMutex;

static GtkWidget *itemPopup = NULL;

void tlDrawEntry(tlType *tl, tocEntry *entry, gint *firstTrack);
void tlDrawRuler(tlType *tl);
void tlDrawRulerLevel(tlType *tl, gdouble spacing, glong height, gboolean drawLabels);
void updateStatusBar(tlType *tl, glong time);


void tlConfigure(tlType *tl)
{
    if (tl) {
        if (tl->pixmap)
            gdk_pixmap_unref(tl->pixmap);
    
        tl->pixmap = gdk_pixmap_new(tl->timeline->window,
            tl->timeline->allocation.width,
            tl->timeline->allocation.height,
            -1);
        tlSetScroller(tl, -1);
        redrawTimeline(tl);
    }
}


void tlExpose(tlType *tl, GdkEventExpose  *event)
{
	if (tl->pixmap) {
		gdk_draw_pixmap(tl->timeline->window,
		    tl->timeline->style->fg_gc[GTK_STATE_NORMAL],
    	    tl->pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
	}
    //tl->cursor.lastDrawnTime_S = -1;
    tlDrawCursor(tl);
}


void redrawTimeline(tlType *tl)
{
	GList       *item;
	tocEntry    *entry;
    gint        nextBreak;

    if (!tl->pixmap) return;
    
    tl->tlEndTime = tlPointToTime(tl, tl->timeline->allocation.width-1);

    if (tl->timeline->allocation.width < tlTimeToPoint(tl, 
                                    prefs.maxCdLength*60*FRAMES_PER_SEC)) {
        gdk_draw_rectangle(tl->pixmap,
            tl->timeline->style->bg_gc[GTK_STATE_NORMAL],
            TRUE,
            0, 0,
            tl->timeline->allocation.width,
            tl->timeline->allocation.height);
    } else {
        gdk_draw_rectangle(tl->pixmap,
            tl->timeline->style->bg_gc[GTK_STATE_NORMAL],
            TRUE,
            0, 0,
            tlTimeToPoint(tl, prefs.maxCdLength*60*FRAMES_PER_SEC)+1,
            tl->timeline->allocation.height);
        gdk_draw_rectangle(tl->pixmap,
            tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
            TRUE,
            tlTimeToPoint(tl, prefs.maxCdLength*60*FRAMES_PER_SEC)+1, 0,
            tl->timeline->allocation.width - tlTimeToPoint(tl, prefs.maxCdLength*60*FRAMES_PER_SEC),
            tl->timeline->allocation.height);
    }
    
    tlDrawRuler(tl);

    nextBreak = 1;
    if (tl->document) {
        for(item=tl->document->entries; item; item=g_list_next(item)) {
            entry = item->data;
            if(entry->start_F + entry->audioLength_F + entry->pregapLength_F >=
                tl->tlStart && entry->start_F <= tl->tlEndTime)
                tlDrawEntry(tl, entry, &nextBreak);
            else
                nextBreak += tocEntryCountBreaks(entry);
        }
    }

    gtk_widget_draw(tl->timeline, NULL);
    tl->cursor.lastDrawnTime_S = -1;
    tl->cursor.oldLastClickFrame = 0;
    tlDrawCursor(tl);
}


void tlDrawRuler(tlType *tl)
{
    gfloat tickWidth;
    gint i;
    
    i = 0;
    while (rulerSize[i+1]) {
        tickWidth = tl->timeline->allocation.width /(gfloat)(tl->tlEndTime - tl->tlStart) * rulerSize[i];
        if (tickWidth > 100)
            break;
        i++;
    }
    tlDrawRulerLevel(tl, rulerSize[i], 24, TRUE);
    tlDrawRulerLevel(tl, rulerSize[i]/2.0, 12, FALSE);
    tlDrawRulerLevel(tl, rulerSize[i]/5.0, 8, FALSE);
    tlDrawRulerLevel(tl, rulerSize[i]/10.0, 4, FALSE);
}


void tlDrawRulerLevel(tlType *tl, gdouble spacing, glong height, gboolean drawLabels)
{
    gint  i;
    gchar timeLabel[18];
    PangoLayout *layout;
    
    if (drawLabels)
        layout = gtk_widget_create_pango_layout(tl->timeline, "");

    for (i = tl->tlStart/spacing; 
         i <= MIN(tl->tlEndTime, prefs.maxCdLength*60*FRAMES_PER_SEC)/spacing;
         i++) {
        gdk_draw_line(tl->pixmap, tl->timeline->style->fg_gc[GTK_STATE_ACTIVE],
            tlTimeToPoint(tl, i*spacing+0.5), 0,
            tlTimeToPoint(tl, i*spacing+0.5), height);
        if (drawLabels) {
            gintToCdTime(timeLabel, (glong)(i*spacing+0.5));
            pango_layout_set_text(layout, timeLabel, -1);
            gdk_draw_layout(tl->pixmap, tl->timeline->style->fg_gc[GTK_STATE_ACTIVE], 
                            tlTimeToPoint(tl, i*spacing+0.5) + 5, height - 12,
                            layout);
        }
    }
    if (drawLabels)
        g_object_unref(layout);
}


void tlDrawEntry(tlType *tl, tocEntry *entry, gint *nextBreak)
{
    glong   start, end;
    glong   height;
    glong   breakLoc;
    glong   i;
    glong   wavStart; // Pregap
    glong   wavEnd;
    gchar   breakStr[8];
    GList   *breaks;
    PangoLayout *layout;
    unsigned char lmin, lmax, rmin, rmax;
    
    start = entry->start_F;
    if (start < tl->tlStart) start = tl->tlStart;
    start = tlTimeToPoint(tl, start);

    wavStart = entry->start_F + entry->pregapLength_F;
    if (wavStart < tl->tlStart) wavStart = tl->tlStart;
    wavStart = tlTimeToPoint(tl, wavStart);

    end = entry->start_F + tocEntryLength(entry);
    if (entry->thumbnail) {
        wavEnd = MIN(end, entry->start_F + entry->pregapLength_F -
                     entry->offset_F + entry->thumbnail->length/147);
    } else {
        wavEnd = end;
    }
    end = tlTimeToPoint(tl, end);
    wavEnd = tlTimeToPoint(tl, wavEnd);
    if (end > tl->timeline->allocation.width) end = tl->timeline->allocation.width;
    if (wavEnd > tl->timeline->allocation.width) wavEnd = tl->timeline->allocation.width;
    
    // White background for audio
    gdk_draw_rectangle(tl->pixmap,
        tl->timeline->style->fg_gc[GTK_STATE_SELECTED],
        TRUE,
        start, TL_ENTRY_TOP,
        end - start,
        tl->timeline->allocation.height - TL_ENTRY_TOP - TL_ENTRY_BOTTOM - 1);

    // Show waveform
    height = tl->timeline->allocation.height - TL_ENTRY_TOP - TL_ENTRY_BOTTOM - 1;
    if (entry->fileType == teAudio) {
        g_mutex_lock(thumbMutex);
        if (entry->thumbnail && entry->thumbnail->status != thumbNone) {
            for (i=wavStart; i<wavEnd; i++) {
                thumbnailGetRange(entry, tlPointToThumb(tl, entry, i), 
                    tlPointToThumb(tl, entry, i+1), &lmin, &lmax, &rmin, &rmax);
                gdk_draw_line(tl->pixmap,  tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
                    i, lmin/512.0 * 
                    height + TL_ENTRY_TOP,
                    i, lmax/512.0 * 
                    height + TL_ENTRY_TOP);
                gdk_draw_line(tl->pixmap,  tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
                    i, rmin/512.0 * 
                    height + TL_ENTRY_TOP + height/2,
                    i, rmax/512.0 * 
                    height + TL_ENTRY_TOP + height/2);
                if (tl->tlZoom < 3) {
                    gdk_draw_line(tl->pixmap,  tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
                        i, entry->thumbnail->data[tlPointToThumb(tl, entry, i)]/512.0 * 
                        height + TL_ENTRY_TOP,
                        i+1, entry->thumbnail->data[tlPointToThumb(tl, entry, i+1)]/512.0 * 
                        height + TL_ENTRY_TOP);
                    gdk_draw_line(tl->pixmap,  tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
                        i, entry->thumbnail->data[tlPointToThumb(tl, entry, i)+1]/512.0 * 
                        height + TL_ENTRY_TOP + height/2,
                        i+1, entry->thumbnail->data[tlPointToThumb(tl, entry, i+1)+1]/512.0 * 
                        height + TL_ENTRY_TOP + height/2);
                }
            }
        }
        g_mutex_unlock(thumbMutex);
    } else if (entry->fileType == teSilence) {
        gdk_draw_line(tl->pixmap,  tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
            start, height/4 + TL_ENTRY_TOP,
            end, height/4 + TL_ENTRY_TOP);
        gdk_draw_line(tl->pixmap,  tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
            start, height*3/4 + TL_ENTRY_TOP,
            end, height*3/4 + TL_ENTRY_TOP);
    }
    
    // Show track name
    if (entry->filename) {
        layout = gtk_widget_create_pango_layout(tl->timeline, g_basename(entry->filename));
        // pango_layout_set_width(layout, (end - start)*PANGO_SCALE);
        // pango_layout_set_ellipsize(layout, PANGO_ELLIPSIZE_END);
        gdk_draw_layout(tl->pixmap, tl->timeline->style->fg_gc[GTK_STATE_ACTIVE], 
                        start + 5, tl->timeline->allocation.height - TL_ENTRY_BOTTOM - 20,
                        layout);
        g_object_unref(layout);
    }
    
    // Black box around all entries
    gdk_draw_rectangle(tl->pixmap,
        tl->timeline->style->fg_gc[GTK_STATE_ACTIVE],
        FALSE,
        start, TL_ENTRY_TOP,
        end - start,
        tl->timeline->allocation.height - TL_ENTRY_TOP - TL_ENTRY_BOTTOM - 1);
    
    // Double thick border for selected entries
    if(entry->selected && end-start > 1) {
        gdk_draw_rectangle(tl->pixmap,
            tl->timeline->style->fg_gc[GTK_STATE_ACTIVE],
            FALSE,
            start + 1, TL_ENTRY_TOP + 1,
            end - start - 2,
            tl->timeline->allocation.height - TL_ENTRY_TOP - TL_ENTRY_BOTTOM - 3);
    }

    // Show track breaks
    breaks = entry->breaks;
    layout = gtk_widget_create_pango_layout(tl->timeline, "1");
    while (breaks) {
        breakLoc = entry->start_F + entry->pregapLength_F + (glong)(breaks->data);
        if (breakLoc > tl->tlEndTime) break;
        if (breakLoc >= tl->tlStart) {
            gdk_draw_pixbuf(tl->pixmap, tl->timeline->style->fg_gc[GTK_STATE_ACTIVE],
                markImg[(glong)breaks->data + entry->start_F + entry->pregapLength_F == tl->breakSelected],
                0,0, tlTimeToPoint(tl, breakLoc) -4, TL_ENTRY_TOP - 5,
                -1,-1, GDK_RGB_DITHER_NONE, 0,0);
            snprintf(breakStr, 7, "%d", *nextBreak);
            pango_layout_set_text(layout, breakStr, -1);
            gdk_draw_layout(tl->pixmap, tl->timeline->style->fg_gc[GTK_STATE_ACTIVE], 
                tlTimeToPoint(tl, breakLoc) + 5,
                TL_ENTRY_TOP + 3,
                layout);
        }
        (*nextBreak)++;
        breaks = g_list_next(breaks);
    }
    g_object_unref(layout);
}


gint tlDrawCursor(gpointer data)
{
    tlType *tl = data;
    glong  time_S = getCursorTime(&tl->cursor);
    gboolean playing = tl->cursor.playingFlag;
        
    if (tlTimeToPoint(tl, S2F(tl->cursor.lastDrawnTime_S)) != tlTimeToPoint(tl, S2F(time_S))) {
        if (playing) {
            gdk_draw_line(tl->timeline->window,
                (tl->cursor.useStaticCursor) ? 
                    tl->timeline->style->bg_gc[GTK_STATE_ACTIVE] :
                    tl->timeline->style->fg_gc[GTK_STATE_ACTIVE],
                tlTimeToPoint(tl, S2F(time_S)), 0,
                tlTimeToPoint(tl, S2F(time_S)), tl->timeline->allocation.height-1);
        }
        if(tl->cursor.lastDrawnTime_S != -1) {
            gdk_draw_pixmap(tl->timeline->window,
                tl->timeline->style->fg_gc[GTK_STATE_NORMAL],
                tl->pixmap,
                tlTimeToPoint(tl, S2F(tl->cursor.lastDrawnTime_S)), 0,
                tlTimeToPoint(tl, S2F(tl->cursor.lastDrawnTime_S)), 0,
                1, tl->timeline->allocation.height);
        }
        tl->cursor.lastDrawnTime_S = time_S;
    }

    if (tlTimeToPoint(tl, tl->cursor.lastClickFrame) != tlTimeToPoint(tl, tl->cursor.oldLastClickFrame)) {
        gdk_draw_pixmap(tl->timeline->window,
            tl->timeline->style->fg_gc[GTK_STATE_NORMAL],
            tl->pixmap,
            tlTimeToPoint(tl, tl->cursor.oldLastClickFrame), 0,
            tlTimeToPoint(tl, tl->cursor.oldLastClickFrame), 0,
            1, tl->timeline->allocation.height);
        tl->cursor.oldLastClickFrame = tl->cursor.lastClickFrame;
    }
    gdk_draw_line(tl->timeline->window,
        (tl->cursor.useStaticCursor || !playing) ? 
            tl->timeline->style->fg_gc[GTK_STATE_ACTIVE] :
            tl->timeline->style->bg_gc[GTK_STATE_ACTIVE],
        tlTimeToPoint(tl, tl->cursor.lastClickFrame), 0,
        tlTimeToPoint(tl, tl->cursor.lastClickFrame), tl->timeline->allocation.height-1);

    if (!playing)
        tl->cursor.lastDrawnTime_S = -1;

    return playing;
}


gboolean tlButtonPress(tlType *tl, GdkEventButton  *event)
{
    tocEntry    *entry;
    tlPart      type;
    glong       newValue;
    glong       closestBreak;
    glong       mouseTime;
    
    mouseTime = tlPointToTime(tl, event->x + ((gfloat)TL_PIXELS_PER_FRAME/tl->tlZoom/2));

    if (event->button == 1) {
        tlGetBreakAtPoint(tl, event->x, event->y, &entry, &closestBreak);
        if (closestBreak > -1) {
            tl->breakSelected = closestBreak;
            redrawTimeline(tl);
            updateStatusBar(tl, closestBreak);
            tl->dragType = 3;
            tl->dragStart = event->x;
            return TRUE;
        }
    }
    tlGetEntryAtPoint(tl, event->x, event->y, &entry, &type);
    
    gtk_widget_grab_focus(tl->timeline);
    if (event->button == 1) {
        tl->breakSelected = -1;
        SetActiveCursorTime(tl, mouseTime); 
        if (entry) {
            if (entry->selected) {
                if (event->state & GDK_CONTROL_MASK)
            	    tocDeselect(tl->document, entry);
            } else {
                if (!(event->state & GDK_CONTROL_MASK))
                    tocClearSelection(tl->document);
        	    tocSelect(tl->document, entry);
            }
            tl->dragType = 2;
            tl->dragStart = mouseTime;
        } else {
            if (type & tlEnd) {
                tocClearSelection(tl->document);
            } else {
                tl->cursor.useStaticCursor = FALSE;
            }
            tl->dragType = 1;
            g_mutex_lock(tl->cursor.mutex);
                setCursorTime(&(tl->cursor), F2S(mouseTime));
            g_mutex_unlock(tl->cursor.mutex);
        }
        updateInspector();
        redrawTimeline(tl);
        updateStatusBar(tl, mouseTime);
        return TRUE;
    }
    else if (event->button == 2) {
        tl->dragType = 1;
        g_mutex_lock(tl->cursor.mutex);
            setCursorTime(&(tl->cursor), F2S(mouseTime));
        g_mutex_unlock(tl->cursor.mutex);
        SetActiveCursorTime(tl, mouseTime);
        redrawTimeline(tl);
        updateStatusBar(tl, mouseTime);
        return TRUE;
    }
    else if (event->button == 3) {
        doPopup(tl, event, type, entry);
        return TRUE;
    }
    else if (event->button == 4) {
        newValue = tl->scrollerAdj->value - tl->scrollerAdj->step_increment;
        if (newValue < 0) newValue = 0;
        tlSetScroller(tl, newValue);
        redrawTimeline(tl);
        return TRUE;
    }
    else if (event->button == 5) {
        newValue = tl->scrollerAdj->value + tl->scrollerAdj->step_increment;
        if (newValue > tl->scrollerAdj->upper - tl->scrollerAdj->page_size)
                newValue = tl->scrollerAdj->upper - tl->scrollerAdj->page_size;
        tlSetScroller(tl, newValue);
        redrawTimeline(tl);
        return TRUE;
    }
    return FALSE;
}


gboolean tlMotionNotify(tlType *tl, GdkEventMotion  *event)
{
    tlPart      type;
    tocEntry    *startEntry, *endEntry;
    glong       mouseTime;
    
    if (event->x < 0)
        return FALSE;
    
    if (tl->dragType > 0)
        mouseTime = tlPointToTime(tl, event->x + ((gfloat)TL_PIXELS_PER_FRAME/tl->tlZoom/2));

    if (tl->dragType == 1) {
        g_mutex_lock(tl->cursor.mutex);
            setCursorTime(&(tl->cursor), F2S(mouseTime));
        g_mutex_unlock(tl->cursor.mutex);
        SetActiveCursorTime(tl, mouseTime);
        tl->cursor.useStaticCursor = FALSE;
        tlDrawCursor(tl);
        updateStatusBar(tl, mouseTime);
    } else if (tl->dragType == 2) {
        if (tl->dragStart == -1)
            return FALSE;
        
        startEntry = tocDocWhatIsAtTime(tl->document, tl->dragStart);
        tlGetEntryAtPoint(tl, event->x, -1, &endEntry, &type);

        endEntry = (!(type & tlFirstHalf)) ? tocEntryNext(endEntry): endEntry;

        if (startEntry == endEntry || tocEntryNext(startEntry) == endEntry)
            return FALSE;
        
        if (tl->document->selection) {
            g_mutex_lock(tl->cursor.mutex);
                tocDocMoveSelectionBefore(tl->document, endEntry);
                tl->cursor.changedFlag = 1;
            g_mutex_unlock(tl->cursor.mutex);
            tocDocSetDirty(tl->document, TRUE);
            
            tl->dragStart = mouseTime;

            updateInspector();
            redrawTimeline(tl);
            return TRUE;
        }
    } else if (tl->dragType == 3) {
        if (event->x - tl->dragStart > 4 || tl->dragStart - event->x > 4)
            tl->dragType = 4;
            return TRUE;
    } else if (tl->dragType == 4) {
        if (mouseTime != tocDocFindClosestBreak(
                tl->document, mouseTime)) {
            tocDocRemoveBreak(tl->document, tl->breakSelected);
            tocDocAddBreak(tl->document, mouseTime);
            tl->breakSelected = mouseTime;
            redrawTimeline(tl);
            return TRUE;
        }
    }
    return TRUE;
}


gboolean tlButtonRelease(tlType *tl, GdkEventButton  *event)
{
    tocEntry    *entry;
    glong       closestBreak;

    if (tl->dragType == 3) {
        tlGetBreakAtPoint(tl, event->x, event->y, &entry, &closestBreak);
        if (closestBreak > -1) {
            tocClearSelection(tl->document);
            g_mutex_lock(tl->cursor.mutex);
                setCursorTime(&(tl->cursor), F2S(closestBreak));
            g_mutex_unlock(tl->cursor.mutex);
            tl->breakSelected = closestBreak;
            SetActiveCursorTime(tl, closestBreak);
            updateInspector();
            redrawTimeline(tl);
            updateStatusBar(tl, closestBreak);
        }
    } else if (tl->dragType == 4) {
        SetActiveCursorTime(tl, tlPointToTime(tl, event->x +
                            ((gfloat)TL_PIXELS_PER_FRAME/tl->tlZoom/2))); 
        tlDrawCursor(tl);
        tocDocSetDirty(tl->document, TRUE);
    }
    tl->dragType = 0;
    return FALSE;
}


void updateStatusBar(tlType *tl, glong time) {
    gchar       timeStr[32];

    gintToCdTime(timeStr, time);
    gtk_statusbar_pop(statusBar, statusBarID);
    gtk_statusbar_push(statusBar, statusBarID, timeStr);
}


gboolean tlEnterNotify(tlType *tl, GdkEventCrossing *event)
{
    return FALSE;
}
    

gboolean tlLeaveNotify(tlType *tl, GdkEventCrossing *event)
{
    //gtk_statusbar_pop(statusBar, statusBarID);
    return TRUE;
}


void tlGetBreakAtPoint(tlType *tl, gint x, gint y, tocEntry **retEntry, glong *retBreak)
{
    glong   time, closest;
    
    if (retEntry) *retEntry = NULL;
    *retBreak = -1;

    if (!(y >= TL_ENTRY_TOP - 5 && y <= TL_ENTRY_TOP + 11)) 
        return;
    
    time = tlPointToTime(tl, x);

    closest = tocDocFindClosestBreak(tl->document, time);
    if (closest == -1 || tlTimeToPoint(tl, closest) - x > 4 ||
                         x - tlTimeToPoint(tl, closest) > 4)
        return;
    
    *retBreak = closest;
    if (retEntry) *retEntry = tocDocWhatIsAtTime(tl->document, closest);

    return;
}


void tlGetEntryAtPoint(tlType *tl, gint x, gint y, tocEntry **retEntry, tlPart *retType)
{
    glong time;
    
    *retType = tlNone;
    *retEntry = NULL;
    if (y != -1 && !(y >= TL_ENTRY_TOP && y <= tl->timeline->allocation.height - TL_ENTRY_BOTTOM - 1)) 
        return;
    
    time = tlPointToTime(tl, x);
    *retEntry = tocDocWhatIsAtTime(tl->document, time);
    
    if (!*retEntry) {
        *retType += tlEnd;
        return;
    }
    
    if ((*retEntry)->fileLength_F > -1)
        *retType += tlEntry;
    else
        *retType += tlSilence;

    if (x < tlTimeToPoint(tl, (*retEntry)->start_F) + 3 ||
        x >= tlTimeToPoint(tl, (*retEntry)->start_F) + tlTimeToPoint(tl, (*retEntry)->audioLength_F) - 3) {
            *retType += tlEdge;
    }
    
    if (time < (*retEntry)->start_F + (*retEntry)->audioLength_F/2) {
            *retType += tlFirstHalf;
    }
}


gboolean tlKeyPress(tlType *tl, GdkEventKey *event)
{
    glong newValue;

    switch(event->keyval){
        case GDK_Delete:
            if (tl->document->selection) {
                g_mutex_lock(tl->cursor.mutex);
                    tocDocDeleteSelection(tl->document);
                    tl->cursor.changedFlag = 1;
                g_mutex_unlock(tl->cursor.mutex);
                tocDocSetDirty(tl->document, TRUE);
                updateInspector();
                tlSetScroller(tl, -1);
                redrawTimeline(tl);
                return TRUE;
            } else if (tl->breakSelected != -1) {
                tocDocRemoveBreak(tl->document, tl->breakSelected);
                tocDocSetDirty(tl->document, TRUE);
                redrawTimeline(tl);
                return TRUE;
            }
            break;
        case GDK_Up:
            tlZoomIn(tl);
            redrawTimeline(tl);
            return TRUE;
        case GDK_Down:
            tlZoomOut(tl);
            redrawTimeline(tl);
            return TRUE;
        case GDK_equal:
            tlZoomFit(tl);
            redrawTimeline(tl);
            return TRUE;
        case GDK_Left:
            if (event->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK))
                newValue = tl->scrollerAdj->value - tl->scrollerAdj->page_increment;
            else
                newValue = tl->scrollerAdj->value - tl->scrollerAdj->step_increment;

            if (newValue < 0) newValue = 0;
            tlSetScroller(tl, newValue);
            redrawTimeline(tl);
            return TRUE;
        case GDK_Right:
            if (event->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK))
                newValue = tl->scrollerAdj->value + tl->scrollerAdj->page_increment;
            else
                newValue = tl->scrollerAdj->value + tl->scrollerAdj->step_increment;

            if (newValue > tl->scrollerAdj->upper - tl->scrollerAdj->page_size)
                    newValue = tl->scrollerAdj->upper - tl->scrollerAdj->page_size;
            tlSetScroller(tl, newValue);
            redrawTimeline(tl);
            return TRUE;
        case GDK_s:
            tlSplitAtCursor(tl);
            tocDocSetDirty(tl->document, TRUE);
            return TRUE;
        case GDK_i:
            on_insertFile_activate(NULL, NULL);
            tocDocSetDirty(tl->document, TRUE);
            return TRUE;
        case GDK_b:
            if (tl->breakSelected != -1 && tl->cursor.playingFlag) tl->cursor.useStaticCursor = FALSE;
            newValue = S2F(GetActiveCursorTime(tl));
            tocDocAddBreak(tl->document, newValue);
            tocClearSelection(tl->document);
            tl->breakSelected = newValue;
            redrawTimeline(tl);
            tocDocSetDirty(tl->document, TRUE);
            return TRUE;
        case GDK_space:
            tlInsertSilence(tl);
            tocDocSetDirty(tl->document, TRUE);
            return TRUE;
        case GDK_p:
            on_play_activate(NULL, NULL);
            return TRUE;
        case GDK_t:
            on_stop_activate(NULL, NULL);
            return TRUE;
    }
    return FALSE;
}


void doPopup(tlType *tl, GdkEventButton *event, tlPart type, tocEntry *entry)
{
    if(!itemPopup)
        itemPopup = create_itemPopup();
    if(entry){
        if(!entry->selected) {
            tocClearSelection(tl->document);
            tocSelect(tl->document, entry);
            updateInspector();
            redrawTimeline(tl);
        }
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popPaste"), FALSE);
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popCopy"), TRUE);
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popCut"), TRUE);
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popDelete"), TRUE);
    } else {
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popPaste"), TRUE);
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popCopy"), FALSE);
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popCut"), FALSE);
        gtk_widget_set_sensitive(lookup_widget(itemPopup,"popDelete"), FALSE);
    }
    gtk_menu_popup((GtkMenu*)itemPopup, NULL, NULL, NULL, NULL,
        ((GdkEventButton *)event)->button, ((GdkEventButton *)event)->time);
}

glong tlPointToTime(tlType *tl, glong x)
{
    return tl->tlStart + (x * tl->tlZoom / (gfloat)TL_PIXELS_PER_FRAME);
}


glong tlPointToThumb(tlType *tl, tocEntry *entry, glong x)
{
    return MIN((tl->tlStart + (x * tl->tlZoom / (gfloat)TL_PIXELS_PER_FRAME) - 
               entry->start_F + entry->offset_F - entry->pregapLength_F)*
               ((gfloat)SAMPLES_PER_SEC/FRAMES_PER_SEC)/4,
               entry->thumbnail->length-2);
}


glong tlTimeToPoint(tlType *tl, glong t)
{
    return (t - tl->tlStart) * (gfloat)TL_PIXELS_PER_FRAME / tl->tlZoom;
}


void tlScrollerChanged(tlType *tl)
{
    tl->tlStart = tl->scrollerAdj->value;
    redrawTimeline(tl);
}


/* start = -1 means don't change its value */
void tlSetScroller(tlType *tl, gint start)
{
    if(!tl->scrollerAdj) return;
    
    if(start != -1) {
        tl->tlStart = start;
    }
    tl->scrollerAdj->value = tl->tlStart;
    tl->scrollerAdj->page_size = (tlPointToTime(tl, tl->timeline->allocation.width-1) - tl->tlStart);
    tl->scrollerAdj->upper = tocDocLength(tl->document) + 2*FRAMES_PER_SEC;
    if (tl->scrollerAdj->upper < tl->scrollerAdj->page_size)
        tl->scrollerAdj->upper = tl->scrollerAdj->page_size;
    tl->scrollerAdj->step_increment = MAX(1, tl->scrollerAdj->page_size / 20);
    tl->scrollerAdj->page_increment = tl->scrollerAdj->page_size / 1.5;
    gtk_adjustment_changed(tl->scrollerAdj);
}


void tlZoomIn(tlType *tl)
{
    glong width;
    
    if (tl->tlZoom < TL_ZOOM_MIN * TL_ZOOM_INC)
        tl->tlZoom = TL_ZOOM_MIN;
    else
        tl->tlZoom = tl->tlZoom / TL_ZOOM_INC;
    
    width = tlPointToTime(tl, tl->timeline->allocation.width-1) - tlPointToTime(tl, 0);
    tl->tlStart = S2F(GetActiveCursorTime(tl)) - width/2;
    if (tl->tlStart < 0) tl->tlStart = 0;
    
    tlSetScroller(tl, -1);
}


void tlZoomOut(tlType *tl)
{
    glong width;
    
    if (tl->tlZoom > TL_ZOOM_MAX / TL_ZOOM_INC)
        tl->tlZoom = TL_ZOOM_MAX;
    else
        tl->tlZoom = tl->tlZoom * TL_ZOOM_INC;
    
    width = tlPointToTime(tl, tl->timeline->allocation.width-1) - tlPointToTime(tl, 0);
    tl->tlStart = S2F(GetActiveCursorTime(tl)) - width/2;
    if (tl->tlStart < 0) tl->tlStart = 0;
    
    tlSetScroller(tl, -1);
}


void tlZoomFit(tlType *tl)
{
    glong   length;
    
    tl->tlStart = 0;

    length = tocDocLength(tl->document);
    if (length == 0)
        tl->tlZoom = TL_ZOOM_START;
    else
        tl->tlZoom = (gfloat)(length + 2*FRAMES_PER_SEC) / 
                     (tl->timeline->allocation.width-1) * (gfloat)TL_PIXELS_PER_FRAME;
    
    tlSetScroller(tl, -1);
}


void tlInsertFile(tlType *tl, gchar *filename)
{
    tocEntry    *newEntry;
    tocEntry    *oldEntry;
    
    tocClearSelection(tl->document);
    newEntry = tocDocNewEntry(tl->document, filename,
                              0, (prefs.autoPregap) ? 150 : 0, -1, TRUE);
    oldEntry = tocDocWhatIsAtTime(tl->document, S2F(GetActiveCursorTime(tl)));
    g_mutex_lock(tl->cursor.mutex);
        tocSelect(tl->document, tocDocInsertEntryBefore(tl->document, newEntry, oldEntry));
        tl->cursor.changedFlag = TRUE;
        SetActiveCursorTime(tl, newEntry->start_F + tocEntryLength(newEntry));
        tocDocSetDirty(tl->document, TRUE);
    g_mutex_unlock(tl->cursor.mutex);
    
    updateInspector();
    redrawTimeline(tl);
    tlSetScroller(tl, -1);
}


void tlInsertSilence(tlType *tl)
{
    tocEntry    *newEntry;
    tocEntry    *oldEntry;
    
    tocClearSelection(tl->document);
    newEntry = tocDocNewEntry(tl->document, NULL, 0, 0, 750, FALSE);
    oldEntry = tocDocWhatIsAtTime(tl->document, S2F(GetActiveCursorTime(tl)));
    g_mutex_lock(tl->cursor.mutex);
        tocSelect(tl->document, tocDocInsertEntryBefore(tl->document, newEntry, oldEntry));
        tl->cursor.changedFlag = TRUE;
        SetActiveCursorTime(tl, newEntry->start_F + tocEntryLength(newEntry));
        tocDocSetDirty(tl->document, TRUE);
    g_mutex_unlock(tl->cursor.mutex);
    
    updateInspector();
    redrawTimeline(tl);
    tlSetScroller(tl, -1);
}


void tlSplitAtCursor(tlType *tl)
{
    tocEntry *entry;
    
    entry = tocDocWhatIsAtTime(tl->document, S2F(GetActiveCursorTime(tl)));
    if (entry) {
        tocDocSplitEntry(tl->document, entry, S2F(GetActiveCursorTime(tl)));
        tocDocSetDirty(tl->document, TRUE);
        redrawTimeline(tl);
    }
}


void tlDrop(tlType *tl, gint x, gchar *data)
{
    gchar       *urilist;
    gchar       *token, *filename;
    glong       length;
    tocDoc      *doc;
    tocEntry    *entry;
    tocEntry    *newEntry;
    tlPart      type;
    
    if (!data) return;
    
    urilist = g_locale_to_utf8(data, -1, NULL, NULL, NULL);
    
    token = strtok(urilist, "\n\r");
    while(token) {
        filename = g_filename_from_uri(token, NULL, NULL);
        if (filename && strlen(filename)) {
            length = getFileLength(filename);
            
            if (length == teNotSound) {
                doc = openTocDoc(filename);
                if (doc) {
                    if (!tl->document || !tl->document->dirty ||
                        AskMessage("Are you sure you want to Close this file without saving?")) {
                        g_mutex_lock(tl->cursor.mutex);
                            if (tl->document) freeTocDoc(tl->document);
                            tl->document = doc;
                            tl->cursor.changedFlag = 1;
                            tocDocSetDirty(tl->document, FALSE);
                        g_mutex_unlock(tl->cursor.mutex);
                        tlZoomFit(tl);
                        break;
                    } else {
                        freeTocDoc(doc);
                    }
                }
            } else if (length > 0) {
                g_mutex_lock(tl->cursor.mutex);
                tocClearSelection(tl->document);
                tlGetEntryAtPoint(tl, x, -1, &entry, &type);
                if (! (type & tlFirstHalf)) {
                    entry = tocEntryNext(entry);
                }
                newEntry = tocDocNewEntry(tl->document, filename, 0, 
                    (prefs.autoPregap) ? 150 : 0, -1, TRUE);
                tocSelect(tl->document, tocDocInsertEntryBefore(tl->document,
                    newEntry, entry));
                SetActiveCursorTime(tl, newEntry->start_F + tocEntryLength(newEntry));
                g_free(filename);
                tocDocSetDirty(tl->document, TRUE);
                g_mutex_unlock(tl->cursor.mutex);
            } else {
                printf(" --- dnd --- bad sound file: \"%s\"\n", filename);
                g_free(filename);
            }
        }
        token = strtok(NULL, "\n\r");
    }
    updateInspector();
    tlSetScroller(tl, -1);
    redrawTimeline(tl);
    g_free(urilist);
}

void tlCreate(tlType **newTl, GtkWidget *timeline, GtkRange *scroller, tocDoc *doc)
{
    tlType *tl;
    static GtkTargetEntry targets[] = { { "text/uri-list", 0, 0 } };
    
    (*newTl) = g_malloc(sizeof(tlType));
    tl = *newTl;
    tl->document = doc;
    tl->pixmap = NULL;
    tl->timeline = timeline;
    tl->scrollerAdj = gtk_range_get_adjustment(scroller);
    tl->cursor.startPos_S = 0;
    tl->cursor.lastDrawnTime_S = -1;
    tl->cursor.oldLastClickFrame = 0;
    tl->cursor.playingFlag = 0;
    tl->cursor.stopFlag = 0;
    tl->cursor.changedFlag = 1;
    tl->cursor.timer = g_timer_new();
    SetActiveCursorTime(tl, 0);
    tl->cursor.mutex = g_mutex_new();
    tl->tlStart = 0;
    markImg[0] = create_pixbuf("mark.png");
    markImg[1] = create_pixbuf("markSelected.png");
    gtk_widget_realize(tl->timeline); // ZoomFit needs this to happen first
    tlZoomFit(tl);
    tl->dragStart = -1;
    tl->dragType = 0;
    tl->breakSelected = -1;
    gtk_drag_dest_set(tl->timeline, GTK_DEST_DEFAULT_ALL, targets, 1, GDK_ACTION_COPY);
    updateStatusBar(tl, 0);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(lookup_widget(GTK_WIDGET(timeline), "auto_pregap")), prefs.autoPregap);
    tocDocSetDirty(tl->document, tl->document->dirty);
}


gboolean tlDestroy(tlType *tl)
{
    g_mutex_free(tl->cursor.mutex);
    return FALSE;
}


glong GetActiveCursorTime(tlType *tl)
{
    if (tl->cursor.useStaticCursor)
        return F2S(tl->cursor.lastClickFrame);
    else
        return getCursorTime(&tl->cursor);
}


void SetActiveCursorTime(tlType *tl, glong newTime)
{
    tl->cursor.useStaticCursor = TRUE;
    tl->cursor.lastClickFrame = newTime;
}


void ShowErrorMessage(gchar *msg)
{
    GtkMessageDialog *msgDialog;

    msgDialog = GTK_MESSAGE_DIALOG(gtk_message_dialog_new(NULL, 
        GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT, 
        GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, msg));
    gtk_dialog_run(GTK_DIALOG(msgDialog));
    gtk_widget_destroy(GTK_WIDGET(msgDialog));
}


gboolean AskMessage(gchar *msg)
{
    GtkMessageDialog *msgDialog;
    gint ret;
    
    msgDialog = GTK_MESSAGE_DIALOG(gtk_message_dialog_new(NULL, 
        GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT, 
        GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, msg));
    ret = gtk_dialog_run(GTK_DIALOG(msgDialog));
    gtk_widget_destroy(GTK_WIDGET(msgDialog));

    return (ret == -8);
}
