/***************************************************************************
 *            timeline.h
 *
 *  Fri Jul 23 16:24:06 2004
 *  Copyright  2004  benjie
 ****************************************************************************/

#ifndef _TIMELINE_H
#define _TIMELINE_H

typedef struct {
    glong    startPos_S;       // time when audio started playing in samples
    GTimer   *timer;           // Timer of duration since started playing audio
    glong    lastDrawnTime_S;  // where did we last draw the cursor (for undrawing)
    gchar    stopFlag;         // if set to TRUE, the sound thread stops itself
    gchar    changedFlag;      // if set to True, the sound thread restarts itself
    gint     playingFlag;      // TRUE if currently playing sound
    glong    lastClickFrame;   // where did the user last click? (in frames)
    glong    oldLastClickFrame;// where was the last field last drawn? (in frames)
    gboolean useStaticCursor;  // TRUE if the static cursor is active
    GMutex   *mutex;           // mutex around changing the position or the toc
} cursorType;

typedef struct {
    GtkWidget       *timeline;
    GtkAdjustment   *scrollerAdj;
    GdkPixmap       *pixmap;
    tocDoc          *document;
    cursorType      cursor;
    glong           dragType;
    glong           dragStart;
    glong           breakSelected;
    gfloat          tlZoom;
    glong           tlStart;
    glong           tlEndTime;
} tlType;


#define SAMPLES_PER_SEC 44100
#define FRAMES_PER_SEC 75

static glong F2S(glong f) {return f*(SAMPLES_PER_SEC/FRAMES_PER_SEC);}
static glong S2F(glong s) {return s/(SAMPLES_PER_SEC/FRAMES_PER_SEC);}

#define tlNone      0
#define tlEntry     1
#define tlSilence   2
#define tlEdge      4
#define tlFirstHalf 8
#define tlEnd       16
#define tlBreak     32
typedef gint tlPart;

void tlCreate(tlType **newTl, GtkWidget *timeline, GtkRange *scroller, tocDoc *doc);
gboolean tlDestroy(tlType *tl);
void tlConfigure(tlType *tl);
void tlExpose(tlType *tl, GdkEventExpose  *event);
void redrawTimeline(tlType *tl);
gint tlDrawCursor(gpointer tl);

gboolean tlKeyPress(tlType *tl, GdkEventKey *event);
gboolean tlButtonPress(tlType *tl, GdkEventButton *event);
gboolean tlMotionNotify(tlType *tl, GdkEventMotion *event);
gboolean tlButtonRelease(tlType *tl, GdkEventButton *event);
gboolean tlEnterNotify(tlType *tl, GdkEventCrossing *event);
gboolean tlLeaveNotify(tlType *tl, GdkEventCrossing *event);

void doPopup(tlType *tl, GdkEventButton *event, tlPart type, tocEntry *entry);

void tlInsertFile(tlType *tl, gchar *filename);
void tlInsertSilence(tlType *tl);
void tlDrop(tlType *tl, gint x, gchar *data);
void tlSplitAtCursor(tlType *tl);

void tlScrollerChanged(tlType *tl);
void tlSetScroller(tlType *tl, gint start);
void tlZoomIn(tlType *tl);
void tlZoomOut(tlType *tl);
void tlZoomFit(tlType *tl);

void tlGetEntryAtPoint(tlType *tl, gint x, gint y, tocEntry **retEntry, tlPart *retPart);
void tlGetBreakAtPoint(tlType *tl, gint x, gint y, tocEntry **retEntry, glong *retBreak);
glong tlPointToTime(tlType *tl, glong x);
glong tlPointToThumb(tlType *tl, tocEntry *entry, glong x);
glong tlTimeToPoint(tlType *tl, glong t);

void SetActiveCursorTime(tlType *tl, glong time);
glong GetActiveCursorTime(tlType *tl);

void ShowErrorMessage(gchar *msg);
gboolean AskMessage(gchar *msg);

#endif
